import heapq
import sys
import random
import networkx as nx
import math
import string
import time
import scipy as sci
import numpy as np
from scipy.sparse.linalg import eigsh
from cvxopt import matrix, solvers

def create_undirected_graph_IC(filename):
    g=nx.Graph()
    
    f=open(filename,'r')
    lines=f.readlines()
    for line in lines:
        #[node1, node2, weight]=line.split()
        #index1=int(node1)
        #index2=int(node2)
        #weight=eval(weight)
        temp=line.split()
        index1=int(temp[0])
        index2=int(temp[1])
        #weight=0.6
        weight=1.0
        g.add_edge(index1,index2,weight=weight) 
    
    #for n,nbrs in g.adjacency_iter():
        #for nbr, weight in nbrs.items():
            #print("%d,%d,%d" %(n,nbr,weight['weight']))
            #a=1        
    f.close()   
    return g

def grim_lp(edgelist_name, groupname, K):
    '''
    edgelist_name: group filename
    groupname: group filename
    K: k groups
    '''
    #undirectd graph (networkx format)
    print "get graph"
    G_orig=create_undirected_graph_IC(edgelist_name)
    g=G_orig
    
    print "GRIM quadratic programming"
    start_time=time.time()

    #group_dict
    group_dict={}
    #C: group size
    C={}
    f=open(groupname,'r')
    lines=f.readlines()
    f.close()
    for line in lines:
        temp_list=line.split()
        node_str_list=temp_list[1:]
        #node_list=[int(x) for x in node_str_list]

        group_dict[temp_list[0]]=node_str_list
        C[temp_list[0]]=1
    
    #inverted index
    node_dict={}
    #for node in g.nodes():
    #    node_dict[str(node)]=[]
    for key in group_dict.keys():
        node_list=group_dict[key]
        for node in node_list:
            node_dict[str(node)]=key
    num_node=len(node_dict)
    
    # vacc_distribution: sequence to distribute vaccine
    vacc_distribution=[0]*(len(group_dict))

    graph=nx.to_scipy_sparse_matrix(g,nodelist=range(1,(num_node+1)))
    lamda,evecs_large=eigsh(graph,1,which='LM',tol=1E-3)
    print "eigenvalue"
    print lamda
    u=[ evecs_large[i][0] for i in range(len(evecs_large)) ]
    
    print "Get alpha, beta and gamma"
    
    group_num=len(group_dict.keys())
    alpha=[0]*group_num
    beta=[0]*group_num
    gamma=[]
    for i in range(group_num):
        temp=[0]*group_num
        gamma.append(temp)
    #print node_dict 
    node_list=node_dict.keys()
    for node in node_list:
        Cnode=int(node_dict[node])
        node=int(node)
        C[str(Cnode)]=C[str(Cnode)]+1
        for nbr in nx.all_neighbors(g, node):
            C[str(Cnode)]=C[str(Cnode)]+1
            w=g[node][nbr]['weight']
            Cnbr=int(node_dict[str(nbr)])
            if Cnbr==Cnode:
                Csize=C[str(Cnode)]
                if Csize>1:
                    #temp=(w*u[(int(node)-1)]*u[(int(nbr)-1)])/( Csize*(Csize-1) )
                    temp=(w*u[(int(node)-1)]*u[(int(nbr)-1)])/float(Csize)
                    beta[(Cnode-1)]=beta[(Cnode-1)]+temp
            
    num_edge=g.number_of_edges()
    for i in range(group_num):
        group_id=i+1
        beta[i]=num_edge*beta[i]/C[str(group_id)]
        
    print "run LP"
    """
    #P
    temp=[]
    for i in range(group_num):
        temp.append([])
        for j in range(group_num):
            if i==j:
                temp[i].append(beta[i])
            else:
                temp[i].append(gamma[i][j])
        temp[i].append(0.0)
    temp.append([0.0]*(group_num+1))
    P=2*matrix(temp)
    
    #q
    temp=[]
    for i in range(len(alpha)):
        temp.append((-1)*alpha[i])
    temp.append(0.0)
    q=matrix(temp)

    #G
    temp=[]
    for i in range((group_num+1)):
        row=[0.0]*(group_num+1)
        row[i]=-1.0
        temp.append(row)
    G=matrix(temp)

    #h
    h=matrix( ([0.0]*(group_num+1)) )
    #A
    A=matrix(1.0, (1, group_num+1))
    """
    #
    temp=[]
    for item in beta:
        temp.append(-1.0*item)
    c=matrix(temp)
    
    #A
    temp=[]
    for i in range(group_num):
        tt=[1.0]
        aa=[0.0]*group_num
        aa[i]=-1
        tt=tt+aa
        temp.append(tt)
    A=matrix(temp)
    #b
    #b=matrix( 1.0  )
    temp=[float(K)]
    aa=[0.0]*group_num
    temp=temp+aa
    b=matrix(temp)
    
    sol=solvers.lp(c,A,b)
    #sol=solvers.qp(P, q, G, h, A, b)
    #print sol['x']
    #result=sol['x']*K
    result=sol['x']
    print result
    
    for i in range(len(vacc_distribution)):
        vacc_distribution[i]=int(round(result[i]))
    
    end_time=time.time()
    cost_time=end_time-start_time
    '''
    print 'eigendrop'
    for drop in eigendrop:
        print("%f\t" % drop)
    print('\n')
    '''
    return [lamda,vacc_distribution, cost_time]

def main():
    '''
    filename: edgelist 1 2 0.5
    groupname: group_id (start from 1) member1 member2
    '''
    edgelist_name=sys.argv[1]
    groupname=sys.argv[2]
    K=int(sys.argv[3])
    
    [old_lamda,vacc_distribution, cost_time]=grim_lp(edgelist_name, groupname, K)
    print cost_time
    #selected_groups.sort()
    print vacc_distribution
    print "old lambda: %f\n" % (old_lamda)
    
    save_file=edgelist_name+"_eigen_edge_lp_lambda"
    f=open(save_file,'w')
    f.write("%f\n" % old_lamda)
    f.close()
    
    #selected_groups.sort()
    save_file=edgelist_name+"_eigen_edge_lp_group_selected_"+str(K)
    f=open(save_file,'w')
    for i in range(len(vacc_distribution)):
        group=i+1
        budge=vacc_distribution[i]
        for j in range(budge):
            f.write("%d\n" % group)
    f.close()
    '''
    #Nunber based selected nodes
    number_based_groups=top_num_select_groups(groupname,K)
    number_based_groups=[int(x) for x in number_based_groups]
    number_based_groups.sort()
    print number_based_groups
   
    eigendrop=nodes_in_groups(number_based_groups,groupname)
    print len(top_selected_nodes)
    save_file=filename+'_Top_selected_nodes'
    rg.save_selected_node(save_file,top_selected_nodes)
    '''

if __name__=='__main__':
    main()
