import heapq
import sys
import random
import networkx as nx
import math
import string
import time
import scipy as sci
import numpy as np
from scipy.sparse.linalg import eigsh
from cvxopt import matrix, solvers

def create_undirected_graph_IC(filename):
    g=nx.Graph()
    
    f=open(filename,'r')
    lines=f.readlines()
    for line in lines:
        #[node1, node2, weight]=line.split()
        #index1=int(node1)
        #index2=int(node2)
        #weight=eval(weight)
        temp=line.split()
        index1=int(temp[0])
        index2=int(temp[1])
        #weight=0.6
        weight=1.0
        g.add_edge(index1,index2,weight=weight) 
    
    #for n,nbrs in g.adjacency_iter():
        #for nbr, weight in nbrs.items():
            #print("%d,%d,%d" %(n,nbr,weight['weight']))
            #a=1        
    f.close()   
    return g

def sdp(edgelist_name, groupname, K):
    '''
    edgelist_name: group filename
    groupname: group filename
    K: k groups
    '''
    #undirectd graph (networkx format)
    print "get graph"
    G_orig=create_undirected_graph_IC(edgelist_name)
    g=G_orig
    
    print "Edge-Deletion SDP"
    start_time=time.time()

    #group_dict
    group_dict={}
    #C: group size
    C={}
    f=open(groupname,'r')
    lines=f.readlines()
    f.close()
    for line in lines:
        temp_list=line.split()
        node_str_list=temp_list[1:]
        #node_list=[int(x) for x in node_str_list]

        group_dict[(temp_list[0])]=node_str_list
        C[temp_list[0]]=len(node_str_list)
    
    #inverted index
    node_dict={}
    #for node in g.nodes():
    #    node_dict[str(node)]=[]
    for key in group_dict.keys():
        node_list=group_dict[key]
        for node in node_list:
            node_dict[str(node)]=key
    num_node=len(node_dict)
    
    # vacc_distribution: sequence to distribute vaccine
    vacc_distribution=[0]*(len(group_dict))

    graph=nx.to_scipy_sparse_matrix(g,nodelist=range(1,(num_node+1)))
    lamda,evecs_large=eigsh(graph,1,which='LM',tol=1E-3)
    print "eigenvalue"
    print lamda
    u=[ evecs_large[i][0] for i in range(len(evecs_large)) ]
    
    print "Get c, g, h"
    #get c=(1,0,..,0), x=(t,x1,x2,x3,...,...)
    group_num=len(group_dict)
    temp=[0.0]*(group_num+1)
    temp[0]=1.0
    c=matrix(temp)
    
    #get G and h
    # t*0+x_1*C_1+...<=K
    h=[matrix([1.0*K])]
    temp=[]
    temp.append([0.])
    for i in range(len(C.keys())):
        size=C[str(i+1)]*1.0
        temp.append([size])
    G=[ matrix(temp)]

    # x_i<=1
    for i in range(len(C.keys())):
        temp=[]
        h+=[matrix(1.0)]
        for j in range(1+len(C.keys())):
            temp.append([0])
        temp[(i+1)]=[1.0]
        G+=[matrix(temp)]

    # -x_i>=0
    for i in range(len(C.keys())):
        temp=[]
        h+=[matrix(0.0)]
        for j in range(1+len(C.keys())):
            temp.append([0])
        temp[(i+1)]=[-1.0]
        G+=[matrix(temp)]
    
    #get the last one
    M=[]
    
    for i in range(num_node):
        temp=[0.0]*num_node
        M.append(temp)
    for n,nbrs in g.adjacency_iter():
        for nbr, weight in nbrs.items():
            #print num_node
            #print n
            #print nbr
            M[n-1][nbr-1]=-1.0*weight['weight']
            #M[n-1][nbr-1]=-1.0
    h+=[matrix(M)]           
    print "run sdp"
    
    big_G=[]
    t_list=[]
    for i in range(num_node):
        temp=[0.0]*num_node
        temp[i]=-1.0
        t_list.append(temp)
    temp=[]
    for list in t_list:
        temp=temp+list
    big_G.append(temp)
    
    for i in range(len(C.keys())):
        x_i=[]
        for j in range(num_node):
            temp=[0.0]*num_node
            x_i.append(temp)
        node_list=group_dict[str(i+1)]
        #set x_i[node-1][nbr-1] to be -1*g[][]['weight']
        for node in node_list:
            for nei in ((g[int(node)]).keys()):
                x_i[int(node)-1][int(nei)-1]=-1.0*g[int(node)][nei]['weight']
        temp=[]
        for list in x_i:
            temp=temp+list
        big_G.append(temp)
    G+=[matrix(big_G)]
    sol=solvers.sdp(c,Gs=G, hs=h)
    #print sol['x']
    #result=sol['x']*K
    result=sol['x']
    new_lamda=result[0]
    print result
    print "new lambda: %f" % new_lamda
    result=result[1:]
    for i in range(len(C)):
        result[i]=result[i]*C[str(i+1)]
        
    for i in range(len(vacc_distribution)):
        vacc_distribution[i]=int(round(result[i]))
    
    end_time=time.time()
    cost_time=end_time-start_time
    '''
    print 'eigendrop'
    for drop in eigendrop:
        print("%f\t" % drop)
    print('\n')
    '''
    return [new_lamda, lamda, vacc_distribution, cost_time]


def main():
    '''
    filename: edgelist 1 2 0.5
    groupname: group_id (start from 1) member1 member2
    '''
    edgelist_name=sys.argv[1]
    groupname=sys.argv[2]
    K=int(sys.argv[3])
    
    [new_lamda, old_lamda,vacc_distribution, cost_time]=sdp(edgelist_name, groupname, K)
    print vacc_distribution
    print cost_time
    print "new lambda: %f, old lambda: %f\n" % (new_lamda,old_lamda)
    
    save_file=edgelist_name+"_eigen_edge_new_expected_lambda_"+str(K)
    f=open(save_file,'w')
    f.write("%f\n" % new_lamda)
    f.close()
    
    save_file=edgelist_name+"_eigen_edge_lambda"
    f=open(save_file,'w')
    f.write("%f\n" % old_lamda)
    f.close()
    
    #selected_groups.sort()
    save_file=edgelist_name+"_eigen_edge_group_selected_"+str(K)
    f=open(save_file,'w')
    for i in range(len(vacc_distribution)):
        group=i+1
        budge=vacc_distribution[i]
        for j in range(budge):
            f.write("%d\n" % group)
    f.close()
    

if __name__=='__main__':
    main()
